/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with this
 * work for additional information regarding copyright ownership. The ASF
 * licenses this file to You under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the
 * License.  You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

/*
 * Based on mod_dump_io
 *
 * Logging module for Internet Systems Consortium's NMSG binary protocol.
 * (https://sie.isc.org/).
 *
 * Will log the entire HTTP request header and the source/destination IP and
 * port to the specified destination.
 *
 * Optional config:
 *
 * NmsgDestIP <ip>[:port]	UDP destination. Default is localhost:1234.
 *
 * NmsgMaxHeader <bytes>	Max size of the request header. Default is 8K.
 *
 * NmsgNoteName	<string>	The name of a r->note; if specified then a
 * 				datagram will only be sent when this note
 *				is specified.
 *
 * NmsgEnvName <string>		As above - but an environment variable.
 *
 * The default is none for both of above; i.e. a datagram is always sent.
 *
 * NmsgModuleDir <path>		Location of the nmsg modules. Default is
 * 				/usr/local/lib/nmsg.
 *
 * NmsgHTTPType <int>		Integer with the HTTP message type. Default is
 * 				LOG_NMSG__ISC__HTTP_TYPE__sinkhole (1).
 *
 * NmsgCompress <yes/no>	Whether to compress or not. Default is no.
 *
 * NmsgMTU <bytes>		Maximum size of an NMSG packet. Default is 1280,
 *				can be increased for loopback or jumbo ethernet.
 *
 * Todo: - not sure about all the livetimes and behavour multithreaded rather
 * than forked. - not checked the IP address translations to packet and IPv6.
 *
 * Contact: dirkx(@)apache.org 1.0.2
 * Contact: edmonds(@)isc.org 1.0.3
 * 1.0.0 	20090403	First release
 * 1.0.1 	20090403	Plow through brigade - even when done - as otherwise
 * 				apache missed the remainder of the brigade. Fix config
 *				reading as well.
 * 1.0.2	20090403	build in strnstr() function from FreeBSD
 * 1.0.3	20090403	make thread-safe, unmap IPv4-mapped IPv6 addresses,
 *				terminate HTTP request buffer with a newline,
 *				cosmetic changes
 * 1.1		20090520	nmsg 0.3 compatibility
 */

#include "httpd.h"
#include "http_connection.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"

#include <nmsg.h>
#include <nmsg/isc/nmsgpb_isc_http.h>

#define LOG_NMSG_MODULE_DIR		"/usr/local/lib/nmsg"
#define LOG_NMSG_DEFAULT_DEST_IP	"127.0.0.1"
#define LOG_NMSG_DEFAULT_PORT		8430
#define LOG_NMSG_DEFAULT_MAXSIZE	(8*1024)
#define LOG_NMSG_DEFAULT_MTU		1280
#define LOG_NMSG_DEFAULT_ISC_HTTP_TYPE	(NMSG__ISC__HTTP_TYPE__sinkhole)

module AP_MODULE_DECLARE_DATA log_nmsg_module;

/* config record */
typedef struct log_nmsg_conf_t {
	/* Configured data */
	char		*note;		/* note to trigger logging on */
	char		*env;		/* env-var to trigger logging on (e.g.
					   set by mod_rewrite) */
	int		maxsize;	/* max size of the request header;
					   trunc when beyond this */
	char		*dst;		/* string with IP address of NMSG
					   receiver */
	int		port;		/* UDP port of the NMSG receiver */
	char		*dir;		/* Directory with libnmsg modules */
	int		type;		/* Message type (ISC assigned) */
	int		compress;	/* Compress with zlib (yes/no) */
	int		mtu;		/* MTU for UDP */

	/* Filled out on log-start - various persistent nmsg objects */
	nmsg_output_t	output;
	nmsg_pbmod_t	mod;
	nmsg_pbmodset_t	ms;
	void		*clos;

	apr_thread_mutex_t  *lock_nmsg;
} log_nmsg_conf_t;

/* private record to buffer the header into */
typedef struct log_nmsg_priv_t {
	char		*buf;
	int		len, truncated, seen;
	log_nmsg_conf_t	*cfg;
} log_nmsg_priv_t;

/* The following routine is lifted from FreeBSD and slightly altered.
 *
 * Copyright (c) 2001 Mike Barcroft <mike (at) FreeBSD.org>
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Chris Torek.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

static const char *
_strnstr(const char *s, const char *find, size_t slen)
{
	char c, sc;
	size_t len;

	if ((c = *find++) != '\0') {
		len = strlen(find);
		do {
			do {
				if (slen < 1 || (sc = *s) == '\0')
					return (NULL);
				--slen;
				++s;
			} while (sc != c);
			if (len > slen)
				return (NULL);
		} while (strncmp(s, find, len) != 0);
		s--;
	}
	return (s);
}

/* End of BSD code */

/*
 * listen to the inbound data; and copy into my private buffer until it seems
 * that we've hit the end and/or we hit some max.
 */
static int
log_nmsg_input_filter(ap_filter_t *f, apr_bucket_brigade *bb,
		      ap_input_mode_t mode, apr_read_type_e block,
		      apr_off_t readbytes)
{
	apr_bucket	*b;
	apr_status_t	ret;
	conn_rec	*c = f->c;
	log_nmsg_priv_t	*priv = f->ctx;
	log_nmsg_conf_t	*ptr = priv->cfg;

	/* Hmm - seems you *have* to plow through it - can we somehow punt ? */
	ret = ap_get_brigade(f->next, bb, mode, block, readbytes);
	if (ret != APR_SUCCESS)
		return (APR_SUCCESS);

	if (priv->seen || priv->len >= ptr->maxsize - 1)
		return (APR_SUCCESS);

	for (b = APR_BRIGADE_FIRST(bb);
	     b != APR_BRIGADE_SENTINEL(bb);
	     b = APR_BUCKET_NEXT(b))
	{
		if (!(APR_BUCKET_IS_METADATA(b))) {
			const char *buf;
			apr_size_t nbytes;
			apr_status_t rv;
			
			rv = apr_bucket_read(b, &buf, &nbytes, APR_BLOCK_READ);
			if (rv != APR_SUCCESS)
				continue;
			if (priv->len + nbytes > ptr->maxsize) {
				nbytes = ptr->maxsize - priv->len;
				priv->truncated = 1;
			}
			if (nbytes && !priv->seen) {
				char *p;
				memcpy(priv->buf + priv->len, buf, nbytes);
				priv->len += nbytes;
				if ((p = (char *) _strnstr(priv->buf, "\n\r\n\r", priv->len)) ||
				    (p = (char *) _strnstr(priv->buf, "\n\n", priv->len)) ||
				    (p = (char *) _strnstr(priv->buf, "\r\n\r\n", priv->len)))
				{
					*p = '\0';
					priv->len = strlen(priv->buf);
					priv->seen = 1;
				}
			}
		}
	}
	return (APR_SUCCESS);
}

/*
 * Reset the various buffers - and install a input filter to capture the
 * header. This should only be done once; on setting up the socket.
 */
static int
log_nmsg_pre_conn(conn_rec *c, void *csd)
{
	/* Limitation - not dealing with keep-alive */
	log_nmsg_priv_t *priv;
	log_nmsg_conf_t *ptr;

	ptr = (log_nmsg_conf_t *) ap_get_module_config(c->base_server->module_config,
						       &log_nmsg_module);

	/* create and reset our private space */
	priv = apr_palloc(c->pool, sizeof(log_nmsg_priv_t));
	priv->buf = apr_palloc(c->pool, ptr->maxsize + 1);
	priv->len = priv->truncated = priv->seen = 0;
	priv->cfg = ptr;

	ap_set_module_config(c->conn_config, &log_nmsg_module, priv);

	ap_add_input_filter("LOG_NMSG_IN", priv, NULL, c);
	return (OK);
}

/*
 * Check that we can parse the IP and ports, connect it all and that is about
 * it.
 */
static int
log_nmsg_post_config(apr_pool_t *pconf, apr_pool_t *plog,
		     apr_pool_t *ptemp, server_rec *s)
{
	struct sockaddr_in nmsg_sock;
	char *p;
	int nmsg_fd;
	int skip = 0;
	log_nmsg_conf_t *cfg;
	nmsg_res res;
	
	cfg = (log_nmsg_conf_t *) ap_get_module_config(s->module_config,
						       &log_nmsg_module);
	cfg->port = LOG_NMSG_DEFAULT_PORT;
	if ((p = rindex(cfg->dst, ':'))) {
		*p++ = '\0';
		cfg->port = atoi(p);
	}

	/* nmsg socket */
	if (inet_pton(AF_INET, cfg->dst, &nmsg_sock.sin_addr)) {
		nmsg_sock.sin_family = AF_INET;
		nmsg_sock.sin_port = htons(cfg->port);
	} else {
		ap_log_error(APLOG_MARK, APLOG_ERR, errno, NULL,
			     "inet_pton failed for '%s'", cfg->dst);
		return (DECLINED);
	}

	nmsg_fd = socket(PF_INET, SOCK_DGRAM, 0);
	if (nmsg_fd < 0) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "socket failed");
		return (DECLINED);
	}

	if (connect(nmsg_fd, (struct sockaddr *) &nmsg_sock,
		    sizeof(struct sockaddr_in)) < 0)
	{
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "connect failed");
		return (DECLINED);
	}

	/* nmsg output buf */
	cfg->output = nmsg_output_open_sock(nmsg_fd, cfg->mtu);
	if (cfg->output == NULL) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
			     "unable to nmsg_output_open_sock()");
		return (DECLINED);
	}

	/* configure compression */
	nmsg_output_set_zlibout(cfg->output, cfg->compress);

	/* nmsg modules */
	cfg->ms = nmsg_pbmodset_init(cfg->dir, 0);
	if (cfg->ms == NULL) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
			     "unable to nmsg_pbmodset_init()");
		return (DECLINED);
	}

	/* http pbnmsg module */
	cfg->mod = nmsg_pbmodset_lookup(cfg->ms, NMSG_VENDOR_ISC_ID,
					MSGTYPE_HTTP_ID);
	if (cfg->mod == NULL) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
			     "unable to acquire module handle");
		return (DECLINED);
	}
	res = nmsg_pbmod_init(cfg->mod, &(cfg->clos));
	if (res != nmsg_res_success) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "unable to initialize module");
		return (DECLINED);
	}
	ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
		     "NMSG messaging configured to sent to %s:%d (mtu %d%s)",
		     cfg->dst, cfg->port, cfg->mtu,
		     cfg->compress ? ", compressed" : "");

	/* nmsg lock */
	apr_thread_mutex_create(&cfg->lock_nmsg, APR_THREAD_MUTEX_UNNESTED, plog);

	return (DECLINED);
}

static int
is_ipv6_v4_mapped(uint8_t *addr) {
	static uint8_t ipv6_v4_mapped[] =
		{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF };
	int i;

	for (i = 0; i < sizeof(ipv6_v4_mapped); i++)
		if (addr[i] != ipv6_v4_mapped[i])
			return (0);

	return (1);
}

/* Build up an NMSG payload and transmit it */
static int
log_nmsg_log_transaction(request_rec *r)
{
	Nmsg__Isc__Http http;
	Nmsg__NmsgPayload *np;
	int m = 0;
	nmsg_res res;
	struct timespec ts;

	log_nmsg_conf_t *cfg = (log_nmsg_conf_t *) ap_get_module_config(
				r->server->module_config, &log_nmsg_module);
	log_nmsg_priv_t *priv = (log_nmsg_priv_t *) ap_get_module_config(
				r->connection->conn_config, &log_nmsg_module);

	/*
	 * Trigger on a note or an environment variable (or both) or, if
	 * neither is set, always trigger.
	 */
	if (cfg->note && apr_table_get(r->notes, cfg->note))
		m++;
	if (cfg->env && apr_table_get(r->subprocess_env, cfg->env))
		m++;

	/* We only skip if we have a note or a env; but none was found */
	if ((m == 0) && (cfg->env || cfg->note))
		goto declined;

	memset(&http, 0, sizeof(http));

	res = nmsg_pbmod_message_init(cfg->mod, &http);
	if (res != nmsg_res_success) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL,
			     "unable to initialize http message");
		goto declined;
	}

	http.type = cfg->type;

	http.srcip.data = (uint8_t *) (r->connection->remote_addr->ipaddr_ptr);
	http.srcip.len = r->connection->remote_addr->ipaddr_len;
	if (http.srcip.len == 16 && is_ipv6_v4_mapped(http.srcip.data)) {
		http.srcip.data += 12;
		http.srcip.len = 4;
	}
	http.has_srcip = true;

	http.srcport = /* ntohs */ (r->connection->remote_addr->port);
	http.has_srcport = true;

	http.dstip.data = (uint8_t *) (r->connection->local_addr->ipaddr_ptr);
	http.dstip.len = r->connection->local_addr->ipaddr_len;
	if (http.dstip.len == 16 && is_ipv6_v4_mapped(http.dstip.data)) {
		http.dstip.data += 12;
		http.dstip.len = 4;
	}
	http.has_dstip = true;

	http.dstport = /* ntohs */ (r->connection->local_addr->port);
	http.has_dstport = true;

	/* terminate the end of the buffer with a newline */
	priv->buf[priv->len] = '\n';
	priv->len += 1;

	/* Current tools do not allow binary (yet) and expect \0 terminator. */
	priv->buf[priv->len] = '\0';
	http.request.data = (unsigned char *) priv->buf;
	http.request.len = strlen(priv->buf) + 1;

	http.has_request = true;

	/*
	 * XXX do we set a 'truncated' warning if (priv->truncated)
	 * such as http.is_truncated = 1;
	 */

	/*
	 * r->request_time: 64 bit with number of microseconds since 00:00:00
	 * january 1, 1970 UTC timespec - in seconds and nanoseconds
	 */
	ts.tv_sec = apr_time_sec(r->request_time); /* micro to seconds */
	ts.tv_nsec = apr_time_usec(r->request_time) * 1000; /* micro to nano */

	np = nmsg_payload_from_message(&http, NMSG_VENDOR_ISC_ID,
				       MSGTYPE_HTTP_ID, &ts);
	if (np != NULL) {
		apr_thread_mutex_lock(cfg->lock_nmsg);
		nmsg_output_append(cfg->output, np);
		apr_thread_mutex_unlock(cfg->lock_nmsg);
	}

	ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, r->server, "NMSG payload queued for sending");

declined:
	/*
	 * Spoof - which may or may not work with keep-alive not tested; nor
	 * with chunked, etc
	 */
	priv->len = priv->truncated = priv->seen = 0;
	return (DECLINED);
}

static void *
log_nmsg_create_sconfig(apr_pool_t *p, server_rec *s)
{
	log_nmsg_conf_t *ptr = apr_pcalloc(p, sizeof *ptr);

	ptr->dst = LOG_NMSG_DEFAULT_DEST_IP;
	ptr->maxsize = LOG_NMSG_DEFAULT_MAXSIZE;
	ptr->note = NULL;
	ptr->env = NULL;
	ptr->type = LOG_NMSG_DEFAULT_ISC_HTTP_TYPE;
	ptr->dir = LOG_NMSG_MODULE_DIR;
	ptr->compress = 0;
	ptr->mtu = LOG_NMSG_DEFAULT_MTU;
	return (ptr);
}

/* On child exit - purge any pending data */
static apr_status_t
log_nmsg_child_exit(void *data)
{
	server_rec *s = data;
	log_nmsg_conf_t *cfg = (log_nmsg_conf_t *) ap_get_module_config(
					s->module_config, &log_nmsg_module);

	nmsg_pbmod_fini(cfg->mod, &(cfg->clos));
	nmsg_output_close(&(cfg->output));
	nmsg_pbmodset_destroy(&(cfg->ms));

	ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
		     "NMSG messaging queue purged and stopped");

	apr_thread_mutex_destroy(cfg->lock_nmsg);

	return (APR_SUCCESS);
}

/*
 * Purely trap the init so we can install a handler which will purge on exit.
 */
static void
log_nmsg_child_init(apr_pool_t *p, server_rec *s)
{
	apr_pool_cleanup_register(p, s, log_nmsg_child_exit, log_nmsg_child_exit);
}

static void
log_nmsg_register_hooks(apr_pool_t * p)
{
	ap_register_input_filter("LOG_NMSG_IN", log_nmsg_input_filter,
				 NULL, AP_FTYPE_CONNECTION + 3);
	ap_hook_post_config(log_nmsg_post_config, NULL, NULL, APR_HOOK_LAST);
	ap_hook_child_init(log_nmsg_child_init, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_pre_connection(log_nmsg_pre_conn, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_log_transaction(log_nmsg_log_transaction, NULL, NULL, APR_HOOK_MIDDLE);
}

static const char *
_set_server_string_slot(cmd_parms *cmd, void *dummy, const char *arg)
{
    int offset = (int)(long)cmd->info;
    char *struct_ptr = (char *) ap_get_module_config(cmd->server->module_config,
						     &log_nmsg_module);

    const char *err = ap_check_cmd_context(cmd,
                                           NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);
    if (err != NULL)
        return (err);

    return (ap_set_string_slot(cmd, struct_ptr, arg));
}

static const char *
_set_server_flag_slot(cmd_parms *cmd, void *dummy, int arg)
{
    int offset = (int)(long)cmd->info;
    char *struct_ptr = (char *) ap_get_module_config(cmd->server->module_config,
						     &log_nmsg_module);

    const char *err = ap_check_cmd_context(cmd,
                                           NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);
    if (err != NULL)
        return (err);

    return (ap_set_flag_slot(cmd, struct_ptr, arg));
}
static const char *
_set_server_int_slot(cmd_parms *cmd, void *dummy, const char *arg)
{
    int offset = (int)(long)cmd->info;
    char *struct_ptr = (char *) ap_get_module_config(cmd->server->module_config,
						     &log_nmsg_module);
    const char *err = ap_check_cmd_context(cmd,
                                           NOT_IN_DIR_LOC_FILE|NOT_IN_LIMIT);
    if (err != NULL)
        return (err);

    return (ap_set_int_slot(cmd, struct_ptr, arg));
}

static const command_rec log_nmsg_cmds[] = {
	AP_INIT_TAKE1("NmsgNoteName", _set_server_string_slot,
		      (void *) APR_OFFSETOF(log_nmsg_conf_t, note),
		      RSRC_CONF, "Note field which triggers sending of the NMSG payload - send always when not set"),
	AP_INIT_TAKE1("NmsgEnvName", _set_server_string_slot,
		      (void *) APR_OFFSETOF(log_nmsg_conf_t, env),
		      RSRC_CONF, "Environment variable which triggers sending of the NMSG payload - send always when not set"),
	AP_INIT_TAKE1("NmsgMaxHeader", _set_server_int_slot,
		      (void *) APR_OFFSETOF(log_nmsg_conf_t, maxsize),
		      RSRC_CONF, "Max size of the HTTP headers passed"),
	AP_INIT_TAKE1("NmsgDestIP", _set_server_string_slot,
		      (void *) APR_OFFSETOF(log_nmsg_conf_t, dst),
		      RSRC_CONF, "Destination for the NMSG UDP packets, IP address and optional :port"),
	AP_INIT_TAKE1("NmsgModuleDir", _set_server_string_slot,
		      (void *) APR_OFFSETOF(log_nmsg_conf_t, dir),
		      RSRC_CONF, "Location of libnmsg modules (default is " LOG_NMSG_MODULE_DIR ")"),
	AP_INIT_TAKE1("NmsgHTTPType", _set_server_int_slot,
		      (void *) APR_OFFSETOF(log_nmsg_conf_t, type),
		      RSRC_CONF, "NMSG type number of the HTTP message"),
	AP_INIT_FLAG("NmsgCompress", _set_server_flag_slot,
		     (void *) APR_OFFSETOF(log_nmsg_conf_t, compress),
		     RSRC_CONF, "Transparent zlib compression of transmitted datagrams"),
	AP_INIT_TAKE1("NmsgMTU", _set_server_int_slot,
		      (void *) APR_OFFSETOF(log_nmsg_conf_t, mtu),
		      RSRC_CONF, "MTU for the NMSG message (default is 1280; change to 8000 for loopback)."),
	{ NULL }
};

module AP_MODULE_DECLARE_DATA log_nmsg_module = {
	STANDARD20_MODULE_STUFF,
	NULL,				/* per dir - create */
	NULL,
	log_nmsg_create_sconfig,	/* per server - create */
	NULL,
	log_nmsg_cmds,
	log_nmsg_register_hooks
};
